#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/timers.h"
#include "driver/gpio.h"
#include "esp_wifi.h"
#include "hardwareprofile.h"
#include "system.h"
#include "driver/uart.h"
#include "uart_primary.h"
#include "utility.h"
#include "packet_manager.h"
#include "asynctasks.h"
#include "esp_task_wdt.h"
#include "state.h"


#define MAIN_TAG "Main"
#define ADD         0x0202

#define CHECK_LEN(p, l) \
    if (p.lenght - HEADER_LEN != l){ \
        send_nack(p.sender,ADD, p.command.cmd, WRONG_FORMAT); \
        break; \
    }

void task_main(void *null) {
    static QueueHandle_t       mainqueue, updatequeue;
    static EventGroupHandle_t  updategroup;
    uint8_t             buffer[1024], tmpbuffer[32], buffer2[TOT_INPUTS+TOT_OUTPUTS];
    uint8_t arg1, arg2, numdac, dacstart, dacfinish, partial420;
    packet_t            packet;
    uint16_t            timer1;
    int i, num, alarm, num1;

    mainqueue = xQueueCreate(64, sizeof(packet_t));
    updatequeue = xQueueCreate(1, 0);
    updategroup = xEventGroupCreate();
    esp_wifi_set_mode(WIFI_MODE_NULL);

    init_gpio();

    init_uart_primary(&mainqueue);
    init_async_process_management();
    init_state_sem();

    // Attendi un po' prima di cominciare a rispondere
    vTaskDelay(200/portTICK_PERIOD_MS);
    xQueueReset(mainqueue);

    esp_task_wdt_add(NULL);
    //TODO: considerare la configurazione
    for (;;) {
        esp_task_wdt_reset();

        if (xQueueReceive(mainqueue, (void *)&packet, 10 / portTICK_PERIOD_MS)) {
            //ESP_LOGI(MAIN_TAG, "Received command 0x%0X", packet.command.cmd);
            switch (packet.command.cmd) {
                case SET_OUTPUTS_TIME:
                    CHECK_LEN(packet, TOT_OUTPUTS+2);
                    timer1 = packet.data[TOT_OUTPUTS] << 8 | packet.data[TOT_OUTPUTS+1];
                    for (i = 0; i < TOT_OUTPUTS; i++)
                        buffer[i] = packet.data[TOT_OUTPUTS-1-i];

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //send_nack(packet.sender, ADD, packet.command.cmd, IMPOSSIBLE);
                        //break;
                    }

                    launch_simple_simple_timer_task(timer1, -1, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTn_TIME_AFTER_INPUT:
                    CHECK_LEN(packet, 7);
                    num1 = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    arg1   = packet.data[3];
                    arg2   = packet.data[4];
                    num = packet.data[5];
                    alarm = packet.data[6] >= TOT_INPUTS*8 ? -1 : packet.data[6];

                    memset(buffer, 0, TOT_OUTPUTS);
                    set_single_bitmap(num1, buffer);

                    memset(buffer2, 0, TOT_INPUTS);
                    if (alarm >= 0)
                        set_single_bitmap((int)alarm, buffer2);

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    launch_timer_after_input_task(timer1, num, arg2, arg1, tmpbuffer[arg1] > arg2, buffer2, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTS_TIME_420MA:
                    CHECK_LEN(packet, TOT_OUTPUTS+4);
                    timer1 = packet.data[TOT_OUTPUTS] << 8 | packet.data[TOT_OUTPUTS+1];
                    arg1 = packet.data[TOT_OUTPUTS+2];
                    arg2 = packet.data[TOT_OUTPUTS+3];
                    for (i = 0; i < TOT_OUTPUTS; i++)
                        buffer[i] = packet.data[TOT_OUTPUTS-1-i];

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    launch_simple_timer_task(timer1, -1, arg2, arg1, tmpbuffer[arg1] > arg2, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTS_TIME_420MA_DAC_REACT:
                    CHECK_LEN(packet, TOT_OUTPUTS+8);
                    timer1 = packet.data[TOT_OUTPUTS] << 8 | packet.data[TOT_OUTPUTS+1];
                    arg1 = packet.data[TOT_OUTPUTS+2];
                    arg2 = packet.data[TOT_OUTPUTS+3];
                    numdac = packet.data[TOT_OUTPUTS+4];
                    dacstart = packet.data[TOT_OUTPUTS+5];
                    dacfinish = packet.data[TOT_OUTPUTS+6];
                    partial420 = packet.data[TOT_OUTPUTS+7];
                    for (i = 0; i < TOT_OUTPUTS; i++)
                        buffer[i] = packet.data[TOT_OUTPUTS-1-i];

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    launch_timer_task(timer1, -1, arg2, arg1, tmpbuffer[arg1] > partial420, numdac, dacstart, dacfinish, partial420, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                default:
                    send_nack(packet.sender, ADD, packet.command.cmd, UNKNOWN_CMD);
                    ESP_LOGI(MAIN_TAG, "unknown command: %04X", packet.command.cmd);
                    break;
            }

            //if (packet.data != NULL)
                //free(packet.data);
        }
    }
}

void app_main(void) {
    xTaskCreate(task_main, "MAIN", 8000, NULL, 2, NULL);

    vTaskDelete(NULL);
}