#include <string.h>
#include "state.h"
#include "debounce.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/event_groups.h"
#include "freertos/timers.h"
#include "system.h"
#include "driver/uart.h"
#include "uart_primary.h"
#include "utility.h"
#include "debounce.h"
#include "asynctasks.h"
#include "esp_task_wdt.h"
#include "packet_manager.h"

#define SAMPLES_NUM 10
#define AVERAGE_NUM 1

#define TAG     "state"

static QueueHandle_t  update_queue;

static SemaphoreHandle_t access_state_sem;
static machine_state_t internal_state = {0};

void init_state_sem() {
    access_state_sem = xSemaphoreCreateMutex();
    update_queue = xQueueCreate(256, sizeof(internal_state_message_t));
}


void set_activity_bitmap_output(uint8_t *bitmap) {
    xSemaphoreTake(access_state_sem, WAIT4SEM);
    set_bitmap(internal_state.activitybitmap, bitmap, TOT_OUTPUTS);
    xSemaphoreGive(access_state_sem);
}

void clear_activity_bitmap_output(uint8_t *bitmap) {
    xSemaphoreTake(access_state_sem, WAIT4SEM);
    clear_bitmap(internal_state.activitybitmap, bitmap, TOT_OUTPUTS);
    xSemaphoreGive(access_state_sem);
}

void clear_activity_bitmap_dac(uint8_t bitmap) {
    xSemaphoreTake(access_state_sem, WAIT4SEM);
    clear_bitmap(&internal_state.activitydac, &bitmap, 1);
    xSemaphoreGive(access_state_sem);
}

void set_activity_bitmap_dac(uint8_t bitmap) {
    xSemaphoreTake(access_state_sem, WAIT4SEM);
    set_bitmap(&internal_state.activitydac, &bitmap, 1);
    xSemaphoreGive(access_state_sem);
}

int check_activity_bitmap_output(uint8_t *buffer) {
    int res;
    xSemaphoreTake(access_state_sem, WAIT4SEM);
    res = check_bitmap_intersection(internal_state.activitybitmap, buffer, TOT_OUTPUTS);
    xSemaphoreGive(access_state_sem);
    return res;
}

int check_activity_bitmap_dac(uint8_t buffer) {
    int res;
    xSemaphoreTake(access_state_sem, WAIT4SEM);
    res = check_bitmap_intersection(&internal_state.activitydac, &buffer, 1);
    xSemaphoreGive(access_state_sem);
    return res;
}

//TODO: risolvi per bene questo problema (se e' un problema)
void takeStateSemaphore() {
    xSemaphoreTake(access_state_sem, WAIT4SEM);
}

void giveStateSemaphore() {
    xSemaphoreGive(access_state_sem);
}

void manage_message(internal_state_message_t message) {
    if (message.notvalid)
        return;

    switch(message.type) {
        case STATE_SET_OUTPUTS:
            set_output_state(message.bitmap);
            break;
        case STATE_CLEAR_OUTPUTS:
            clear_output_state(message.bitmap);
            break;
        case STATE_UPDATE_OUTPUTS:
            update_output_state(message.bitmap);
            break;
        case STATE_CLEAR_ACTIVITY_OUTPUTS:
            clear_activity_bitmap_output(message.bitmap);
            break; 
        case STATE_CLEAR_ACTIVITY_DAC:
            clear_activity_bitmap_dac(message.bitmap[0]);
            break; 
        case STATE_SET_ACTIVITY_OUTPUTS:
            set_activity_bitmap_output(message.bitmap);
            break; 
        case STATE_SET_ACTIVITY_DAC:
            set_activity_bitmap_dac(message.bitmap[0]);
            break; 
        case STATE_UPDATE_DAC:
            update_single_dac_state(message.num, message.value);
            break;
    }
}

void send_output_state(internal_state_message_type_t type, uint8_t *bitmap, int notvalid) {
    internal_state_message_t message;
    memcpy(message.bitmap, bitmap, TOT_OUTPUTS);
    message.notvalid = notvalid;
    message.type = type;
    xQueueSend(update_queue, &message, 0);
}

void send_dac_state(int value, int num, int notvalid) {
    internal_state_message_t message;
    message.notvalid = notvalid;
    message.num = num;
    message.value = value;
    message.type = STATE_UPDATE_DAC;
    xQueueSend(update_queue, &message, 0);
}

void send_clear_activity_dac(uint8_t dac, int notvalid) {
    internal_state_message_t message;
    message.notvalid = notvalid;
    message.bitmap[0] = dac;
    message.type = STATE_CLEAR_ACTIVITY_DAC;
    xQueueSend(update_queue, &message, 0);
}