#include "hardwareprofile.h"
#include "driver/gpio.h"
#include "driver/spi_master.h"
#include <string.h>

extern unsigned char revision[];
uint8_t buffer[11];

void init_gpio() {
    gpio_config_t io_conf;
    io_conf.intr_type    = GPIO_PIN_INTR_DISABLE;
    io_conf.mode         = GPIO_MODE_OUTPUT;
    io_conf.pin_bit_mask = GPIO_SEL_15;
    io_conf.pull_down_en = 0;
    io_conf.pull_up_en   = 0;
    gpio_config(&io_conf);

    io_conf.intr_type    = GPIO_PIN_INTR_DISABLE;
    io_conf.mode         = GPIO_MODE_INPUT;
    io_conf.pin_bit_mask = GPIO_SEL_34;
    io_conf.pull_down_en = 0;
    io_conf.pull_up_en   = 0;
    gpio_config(&io_conf);
}



void init_spi() {
    esp_err_t        ret;
    spi_bus_config_t buscfg = {.miso_io_num     = SPI_DIN,
                               .mosi_io_num     = SPI_DOUT,
                               .sclk_io_num     = SPI_CLOCK,
                               .quadwp_io_num   = -1,
                               .quadhd_io_num   = -1,
                               .max_transfer_sz = PARALLEL_LINES * 8 * 4};
    // Initialize the SPI bus
    ret = spi_bus_initialize(HSPI_HOST, &buscfg, 1);
    ESP_ERROR_CHECK(ret);
}

void spi_data(spi_device_handle_t spi, uint8_t *data, int len, gpio_num_t chipselect) {
    esp_err_t         ret;
    spi_transaction_t t;
    memset(&t, 0, sizeof(t));     // Zero out the transaction
    t.length    = 8 * len;        // Command is 8 bits
    t.tx_buffer = data;           // The data is the cmd itself

    if (chipselect != -1) {
        gpio_set_level(chipselect, 0);
        ets_delay_us(1);
    }

    spi_device_acquire_bus(spi, portMAX_DELAY);
    ret = spi_device_polling_transmit(spi, &t);     // Transmit!
    spi_device_release_bus(spi);

    if (chipselect != -1)
        gpio_set_level(chipselect, 1);

    ESP_ERROR_CHECK(ret);
}
